// TODO
/*
 * -Namenskonvertierung beim Packen (!)
 * -Packen Gesamte Diskette
 * -ev. critical section (gleichzeitiges Lesen und Schreiben einer Disk !?!?!)
 */
#include <windows.h>
#include "wcxhead.h"
#include <stdio.h>
#include <time.h>

#define IFDEBUG   if(debugSwitch)

#define BLOCK_INFO  0
#define ADDR_INFO   1

#define FILETYPE_SCRATCHED 0x00
#define FILETYPE_DEL 0x80
#define FILETYPE_SEQ 0x81
#define FILETYPE_PRG 0x82
#define FILETYPE_USR 0x83
#define FILETYPE_REL 0x84

#define IMAGE_TYPE_NO_TYPE   0
#define IMAGE_TYPE_1541_D64  1
#define IMAGE_TYPE_8050_D80  2
#define IMAGE_TYPE_8250_D82  3
#define IMAGE_TYPE_T64       4

// error codes of error information
#define BLOCK_OK 0x01
#define ERROR_20 0x02
#define ERROR_21 0x03
#define ERROR_22 0x04
#define ERROR_23 0x05
#define ERROR_24 0x06
#define ERROR_25 0x07
#define ERROR_26 0x08
#define ERROR_27 0x09
#define ERROR_28 0x0A
#define ERROR_29 0x0B
#define ERROR_74 0x0F

// update on new format addition
#define MAX_SECTORS_IN_DIR_TRACK 100

#pragma pack(1)

typedef struct {
  BYTE nextDirTrack, nextDirSector;
  BYTE fileType;
  BYTE firstFileTrack, firstFileSector;
  BYTE filename [16];
  BYTE unused [9];
  USHORT nrSectors;
} T_DIR_ENTRY;

typedef struct
{
  BYTE TapeDescr[32];
  BYTE Version[2];
  WORD MaxFiles;
  WORD CurrFiles;
  WORD Reserved;
  BYTE UserDescr[24];
} TapeHeader;

typedef struct
{
  BYTE EntryUsed;
  BYTE FileType;
  WORD StartAddr;
  WORD EndAddr;
  BYTE ReservedA[2];
  DWORD TapePos;
  BYTE ReservedB[4];
  BYTE FileName[16];
} TapeEntry;

typedef struct {
  TapeHeader *Header;
  TapeEntry  *Entry;
  BYTE       *Data;
  WORD currFileSlot;
  DWORD tapeSize;
} T_TAPE_STATE;

typedef struct {
  BYTE nextTrack, nextSector, nextEntry;
  BYTE currNextTrack, currNextSector;
  T_DIR_ENTRY *pFiles;
  int fileCounter;
  BYTE nrOfTracks;
  BOOL firstRun;
  int arrCyclicAccessInfo [MAX_SECTORS_IN_DIR_TRACK];        // max number of sectors in dir track
  BYTE firstDirTrack, firstDirSector;
  BYTE firstBAMTrack, firstBAMSector;
  const unsigned (*pSectorAdresses)[2];
  int errorInformation;
  BYTE *errorData;
} T_DISK_STATE;

typedef union {
  T_TAPE_STATE tapeInfo;
  T_DISK_STATE diskInfo;
} T_STATES;

typedef struct {
  BYTE *ArcName;
  int diskType;
  BYTE *diskData;
  int diskDataSize;
  T_STATES state;
} T_MAIN_STATE;

typedef struct {
  BYTE sectorsLeft;
  BYTE BAMOfTrack[3];
} T_BAM_INFO_TRACK;

typedef struct {
  BYTE nextDirTrack, nextDirSector;
  BYTE formatID;
  BYTE reserved1;
  T_BAM_INFO_TRACK BAM_Data [35];
  BYTE diskName [18];
  BYTE diskID [2];
  BYTE fillByte1;
  BYTE formatIDName [2];
  BYTE fillBytes [4];
  BYTE reserved2 [85];
} T_BAM_1541;

#pragma pack()

/*
  Es existieren vier verschiedene Arten von *.D64 Dateien:

  * 683 Sektoren (= 35 Spuren) ohne Fehlerinformation (Dateigre: 174848 Byte = 170.75 KByte)
  * 683 Sektoren (= 35 Spuren) mit Fehlerinformation  (Dateigre: 175531 Byte = 171.42 KByte)
  * 768 Sektoren (= 40 Spuren) ohne Fehlerinformation (Dateigre: 196608 Byte = 192 KByte)
  * 768 Sektoren (= 40 Spuren) mit Fehlerinformation  (Dateigre: 197376 Byte = 192.75 KByte)

  * 802 Sektoren (= 42 Spuren) ohne Fehlerinformation (Dateigre: 205312 Byte = 200.5 KByte)
  * 802 Sektoren (= 42 Spuren) mit Fehlerinformation  (Dateigre: 206114 Byte = 201.28 KByte)
*/

// update on new format addition
#define MAX_SECTORS_PER_TRACK  30       // absolute max nr of sectors per track for supported disk formats
#define MAX_TRACKS             160      // absolute max nr of tracks of supported disk formats
#define MAX_OVERALL_SECTORS    (MAX_SECTORS_PER_TRACK*MAX_TRACKS)      // nr of tracks * max sectors per track

const unsigned sectorAdresses_d64 [40][2] = {
  { 21, 0x00000 }, { 21, 0x01500 }, { 21, 0x02A00 }, { 21, 0x03F00 }, { 21, 0x05400 },
  { 21, 0x06900 }, { 21, 0x07E00 }, { 21, 0x09300 }, { 21, 0x0A800 }, { 21, 0x0BD00 },
  { 21, 0x0D200 }, { 21, 0x0E700 }, { 21, 0x0FC00 }, { 21, 0x11100 }, { 21, 0x12600 },
  { 21, 0x13B00 }, { 21, 0x15000 }, { 19, 0x16500 }, { 19, 0x17800 }, { 19, 0x18B00 },
  { 19, 0x19E00 }, { 19, 0x1B100 }, { 19, 0x1C400 }, { 19, 0x1D700 }, { 18, 0x1EA00 },
  { 18, 0x1FC00 }, { 18, 0x20E00 }, { 18, 0x22000 }, { 18, 0x23200 }, { 18, 0x24400 },
  { 17, 0x25600 }, { 17, 0x26700 }, { 17, 0x27800 }, { 17, 0x28900 }, { 17, 0x29A00 },
  { 17, 0x2AB00 }, { 17, 0x2BC00 }, { 17, 0x2CD00 }, { 17, 0x2DE00 }, { 17, 0x2EF00 }
};

const unsigned sectorAdresses_d80 [77][2] = {
  { 29, 0x00000 }, { 29, 0x01D00 }, { 29, 0x03A00 }, { 29, 0x05700 }, { 29, 0x07400 }, { 29, 0x09100 }, { 29, 0x0AE00 }, { 29, 0x0CB00 }, { 29, 0x0E800 }, { 29, 0x10500 }, { 29, 0x12200 }, { 29, 0x13F00 }, { 29, 0x15C00 }, { 29, 0x17900 }, { 29, 0x19600 }, { 29, 0x1B300 }, { 29, 0x1D000 }, { 29, 0x1ED00 }, { 29, 0x20A00 }, { 29, 0x22700 }, { 29, 0x24400 }, { 29, 0x26100 }, { 29, 0x27E00 }, { 29, 0x29B00 }, { 29, 0x2B800 }, { 29, 0x2D500 }, { 29, 0x2F200 }, { 29, 0x30F00 }, { 29, 0x32C00 },
  { 29, 0x34900 }, { 29, 0x36600 }, { 29, 0x38300 }, { 29, 0x3A000 }, { 29, 0x3BD00 }, { 29, 0x3DA00 }, { 29, 0x3F700 }, { 29, 0x41400 }, { 29, 0x43100 }, { 29, 0x44E00 },
  { 27, 0x46B00 }, { 27, 0x48600 }, { 27, 0x4A100 }, { 27, 0x4BC00 }, { 27, 0x4D700 }, { 27, 0x4F200 }, { 27, 0x50D00 }, { 27, 0x52800 }, { 27, 0x54300 }, { 27, 0x55E00 }, { 27, 0x57900 }, { 27, 0x59400 }, { 27, 0x5AF00 }, { 27, 0x5CA00 },
  { 25, 0x5E500 }, { 25, 0x5FE00 }, { 25, 0x61700 }, { 25, 0x63000 }, { 25, 0x64900 }, { 25, 0x66200 }, { 25, 0x67B00 }, { 25, 0x69400 }, { 25, 0x6AD00 }, { 25, 0x6C600 }, { 25, 0x6DF00 },
  { 23, 0x6F800 }, { 23, 0x70F00 }, { 23, 0x72600 }, { 23, 0x73D00 }, { 23, 0x75400 }, { 23, 0x76B00 }, { 23, 0x78200 }, { 23, 0x79900 }, { 23, 0x7B000 }, { 23, 0x7C700 }, { 23, 0x7DE00 }, { 23, 0x7F500 }, { 23, 0x80C00 }
};

const unsigned sectorAdresses_d82 [154][2] = {
  { 29, 0x00000 }, { 29, 0x01D00 }, { 29, 0x03A00 }, { 29, 0x05700 }, { 29, 0x07400 }, { 29, 0x09100 }, { 29, 0x0AE00 }, { 29, 0x0CB00 }, { 29, 0x0E800 }, { 29, 0x10500 }, { 29, 0x12200 }, { 29, 0x13F00 }, { 29, 0x15C00 }, { 29, 0x17900 }, { 29, 0x19600 }, { 29, 0x1B300 }, { 29, 0x1D000 }, { 29, 0x1ED00 }, { 29, 0x20A00 }, { 29, 0x22700 }, { 29, 0x24400 }, { 29, 0x26100 }, { 29, 0x27E00 }, { 29, 0x29B00 }, { 29, 0x2B800 }, { 29, 0x2D500 }, { 29, 0x2F200 }, { 29, 0x30F00 }, { 29, 0x32C00 },
  { 29, 0x34900 }, { 29, 0x36600 }, { 29, 0x38300 }, { 29, 0x3A000 }, { 29, 0x3BD00 }, { 29, 0x3DA00 }, { 29, 0x3F700 }, { 29, 0x41400 }, { 29, 0x43100 }, { 29, 0x44E00 },
  { 27, 0x46B00 }, { 27, 0x48600 }, { 27, 0x4A100 }, { 27, 0x4BC00 }, { 27, 0x4D700 }, { 27, 0x4F200 }, { 27, 0x50D00 }, { 27, 0x52800 }, { 27, 0x54300 }, { 27, 0x55E00 }, { 27, 0x57900 }, { 27, 0x59400 }, { 27, 0x5AF00 }, { 27, 0x5CA00 },
  { 25, 0x5E500 }, { 25, 0x5FE00 }, { 25, 0x61700 }, { 25, 0x63000 }, { 25, 0x64900 }, { 25, 0x66200 }, { 25, 0x67B00 }, { 25, 0x69400 }, { 25, 0x6AD00 }, { 25, 0x6C600 }, { 25, 0x6DF00 },
  { 23, 0x6F800 }, { 23, 0x70F00 }, { 23, 0x72600 }, { 23, 0x73D00 }, { 23, 0x75400 }, { 23, 0x76B00 }, { 23, 0x78200 }, { 23, 0x79900 }, { 23, 0x7B000 }, { 23, 0x7C700 }, { 23, 0x7DE00 }, { 23, 0x7F500 }, { 23, 0x80C00 },
  { 29, 0x82300 }, { 29, 0x84000 }, { 29, 0x85D00 }, { 29, 0x87A00 }, { 29, 0x89700 }, { 29, 0x8B400 }, { 29, 0x8D100 }, { 29, 0x8EE00 }, { 29, 0x90B00 }, { 29, 0x92800 }, { 29, 0x94500 }, { 29, 0x96200 }, { 29, 0x97F00 }, { 29, 0x99C00 }, { 29, 0x9B900 }, { 29, 0x9D600 }, { 29, 0x9F300 }, { 29, 0xA1000 }, { 29, 0xA2D00 }, { 29, 0xA4A00 }, { 29, 0xA6700 }, { 29, 0xA8400 }, { 29, 0xAA100 }, { 29, 0xABE00 }, { 29, 0xADB00 }, { 29, 0xAF800 }, { 29, 0xB1500 }, { 29, 0xB3200 }, { 29, 0xB4F00 },
  { 29, 0xB6C00 }, { 29, 0xB8900 }, { 29, 0xBA600 }, { 29, 0xBC300 }, { 29, 0xBE000 }, { 29, 0xBFD00 }, { 29, 0xC1A00 }, { 29, 0xC3700 }, { 29, 0xC5400 }, { 29, 0xC7100 },
  { 27, 0xC8E00 }, { 27, 0xCA900 }, { 27, 0xCC400 }, { 27, 0xCDF00 }, { 27, 0xCFA00 }, { 27, 0xD1500 }, { 27, 0xD3000 }, { 27, 0xD4B00 }, { 27, 0xD6600 }, { 27, 0xD8100 }, { 27, 0xD9C00 }, { 27, 0xDB700 }, { 27, 0xDD200 }, { 27, 0xDED00 },
  { 25, 0xE0800 }, { 25, 0xE2100 }, { 25, 0xE3A00 }, { 25, 0xE5300 }, { 25, 0xE6C00 }, { 25, 0xE8500 }, { 25, 0xE9E00 }, { 25, 0xEB700 }, { 25, 0xED000 }, { 25, 0xEE900 }, { 25, 0xF0200 },
  { 23, 0xF1B00 }, { 23, 0xF3200 }, { 23, 0xF4900 }, { 23, 0xF6000 }, { 23, 0xF7700 }, { 23, 0xF8E00 }, { 23, 0xFA500 }, { 23, 0xFBC00 }, { 23, 0xFD300 }, { 23, 0xFEA00 }, { 23, 0x100100 }, { 23, 0x101800 }, { 23, 0x102F00 }
};

const char *fileTypes [] = {
  ".del",
  ".seq",
  ".prg",
  ".usr",
  ".rel"
};

const char *errorText [] = {
  "",                                   // not used
  "00, No error, Sektor ok.",              // N/A
  "20, Header block not found",            // Read
  "21, No sync character",                 // Seek
  "22, Data block not present",            // Read
  "23, Checksum error in data block",      // Read
  "24, Write verify (on format)",          // Write
  "25, Write verify error",                // Write
  "26, Write protect on",                  // Write
  "27, Checksum error in header block",    // Seek
  "28, Write error",                       // Write
  "29, Disk ID mismatch",                  // Seek
  "74, Disk Not Ready (no device 1)"       // Read
};

const char *badChars = "\\/:*?\"<>|";

char iniFile [MAX_PATH];
BOOL debugSwitch=FALSE;

void markSectorAsUsed(T_MAIN_STATE *, BYTE, BYTE, FILE *);

//////////////////////////////////////////////////////////////////////////////
void setNrOfTracks(T_MAIN_STATE *pInfo)
{
  // 40 track image ?
  if (pInfo->diskDataSize >= 768*256)              // 40 tracks w/o error info
  {
    pInfo->state.diskInfo.nrOfTracks = 40;
    if (pInfo->diskDataSize == 768*257)            // 40 tracks with error info
    {
      pInfo->state.diskInfo.errorInformation = TRUE;
      pInfo->state.diskInfo.errorData = pInfo->diskData+pInfo->diskDataSize-768;
    }
    else
      pInfo->state.diskInfo.errorInformation = FALSE;
  }
  else
  {
    // 35 track image
    pInfo->state.diskInfo.nrOfTracks = 35;
    if (pInfo->diskDataSize == 683*257)
    {
      pInfo->state.diskInfo.errorInformation = TRUE;
      pInfo->state.diskInfo.errorData = pInfo->diskData+pInfo->diskDataSize-683;
    }
    else
      pInfo->state.diskInfo.errorInformation = FALSE;
  }
}

//////////////////////////////////////////////////////////////////////////////
int getAbsSectorIndex(T_DISK_STATE *pDiskState, BYTE track, BYTE sector)
{
  int x, index=0;

  if (track < 1 || track > pDiskState->nrOfTracks)
    return (-1);

  if (sector > pDiskState->pSectorAdresses[track-1][BLOCK_INFO])
    return (-1);

  // sector 0 is track 1, sector 0
  for (x=0; x<track-1; x++)
    index += pDiskState->pSectorAdresses[x][BLOCK_INFO];
  index += sector;

  return index;
}

//////////////////////////////////////////////////////////////////////////////
void trimString (char *s)
{
  char *p=s+strlen(s)-1;

  // empty or short s ?
  if (p <= s)
    return;

  while(*p-- == ' ');
  *(p+2) = '\0';
}

//////////////////////////////////////////////////////////////////////////////
void convertPCString(BYTE *filename, unsigned size)
{
  // a -> A
  // A -> a

  int x;
  BYTE c;

  for (x=0; x<size; x++)
  {
    c = filename[x];

    if (c == '[')
      c = 27;
    else
      if (c == ']')
        c = 29;
      else
        if (c >= 'a' && c <= 'z')
          c = c - 'a' + 'A';
        else
          if (c >= 'A' && c <= 'Z')
            c = c - 'A' + 'a';
          else
            if (c == '_')
              c = ' ';
            else
              filename[x] = c;
  }
}

//////////////////////////////////////////////////////////////////////////////
void convertCBMString(BYTE *filename, unsigned size)
{
  int x;
  BYTE c;

  for (x=0; x<size; x++)
  {
    c = filename[x];

    if (c >= 128)                       // 128-255 -> 0-127 (invers ignorieren)
      c -= 128;

    if (c == '\0')
      break;

    if (c >= 'A' && c <= 'Z')           // 65-90 -> 97-122 (A -> a)
      c = c - 'A' + 'a';
    else
      if (c >= 1 && c <= 26)            // 1-26 -> 97-122 (^A -> a)
        c = c - 1 + 'a';

    if (c == 27)                        // 27 -> 90 ([ -> [)
      c = '[';

    if (c == 29)                        // 29 -> 93 (] -> ])
      c = ']';

    if (c == 28 || c == 30 || c == 31)             // 28, 30, 31 -> _
      c = '_';

    // bad char ?
    if (strchr(badChars, c) == NULL)    // "\\/:*?\"<>|"
      filename[x] = c;  // no, take it
    else
      filename[x] = '_';  // yes, replace
  }
}

//////////////////////////////////////////////////////////////////////////////
int GetDiskSector(T_MAIN_STATE *pInfo, BYTE track, BYTE sector, BYTE *sectorBuffer)
{
  int index, errorInfo, rc;
  char msgBuffer [200];

  if (track < 1 || track > pInfo->state.diskInfo.nrOfTracks)
    return (-1);

  if (sector > pInfo->state.diskInfo.pSectorAdresses[track-1][BLOCK_INFO])
    return (-1);

  // check error info
  if (pInfo->state.diskInfo.errorInformation == TRUE)
  {
    // get absolute index of sector
    index = getAbsSectorIndex(&pInfo->state.diskInfo, track, sector);
    errorInfo = pInfo->state.diskInfo.errorData[index];
    if (errorInfo != BLOCK_OK)
    {
      sprintf(msgBuffer, "disk image error occured at %d:%d\n[%s]", track, sector, errorText[errorInfo]);
      rc = MessageBox(NULL, msgBuffer, "DIRCBM", MB_ICONERROR);
      return errorInfo;
    }
  }
  memcpy(sectorBuffer, pInfo->diskData+pInfo->state.diskInfo.pSectorAdresses[track-1][ADDR_INFO]+sector*256, 256);

  return (0);
}

//////////////////////////////////////////////////////////////////////////////
BOOL GetDiskSectorAddr(T_MAIN_STATE *pInfo, BYTE track, BYTE sector, BYTE **addr)
{
  int index, errorInfo, rc;
  char msgBuffer [200];

  if (track < 1 || track > pInfo->state.diskInfo.nrOfTracks)
    return (-1);

  if (sector > pInfo->state.diskInfo.pSectorAdresses[track-1][BLOCK_INFO])
    return (-1);

  // check error info
  if (pInfo->state.diskInfo.errorInformation == TRUE)
  {
    // get absolute index of sector
    index = getAbsSectorIndex(&pInfo->state.diskInfo, track, sector);
    errorInfo = pInfo->state.diskInfo.errorData[index];
    if (errorInfo != BLOCK_OK)
    {
      sprintf(msgBuffer, "disk image error occured at %d:%d\n[%s]", track, sector, errorText[errorInfo]);
      rc = MessageBox(NULL, msgBuffer, "DIRCBM", MB_ICONERROR);
      return errorInfo;
    }
  }
  *addr = pInfo->diskData + pInfo->state.diskInfo.pSectorAdresses[track-1][ADDR_INFO] + sector*256;

  return (0);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
void determineIniAndLogFile(FILE **fdLog, char *logFile)
{
  char path [MAX_PATH];
  int rc;

  // get windows directory
  rc = GetTempPath(MAX_PATH-1, path);
  sprintf(iniFile, "%s\\dircbm.ini", path);
//  sprintf(logFile, "%s\\dircbm_%d.log", buffer, time(NULL));

  rc = GetTempFileName(path, "loo", 0, logFile);

  // get state of logging in ini file if available
  if (*fdLog == NULL)
  {
    debugSwitch = GetPrivateProfileInt("common", "logging", FALSE, "dircbm.ini");

    IFDEBUG *fdLog = fopen (logFile, "wt");
    IFDEBUG setbuf(*fdLog, NULL);
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
int getTypeFromExtension(char *extension, FILE *fdLog)
{
  int x, rc;

  rc = IMAGE_TYPE_NO_TYPE;

  x = atoi(extension+1);
  if ( (*extension == 'd') || (*extension == 'D') )
  {
    switch(x)
    {
      case 64:
        rc = IMAGE_TYPE_1541_D64;
        IFDEBUG fprintf(fdLog, "disk type = D64\n");
        break;

      case 80:
      case 82:
        if (x == 80)
        {
          rc = IMAGE_TYPE_8050_D80;
          IFDEBUG fprintf(fdLog, "disk type = D80\n");
        }
        else
        {
          rc = IMAGE_TYPE_8250_D82;
          IFDEBUG fprintf(fdLog, "disk type = D82\n");
        }
        break;

      default:
        IFDEBUG fprintf(fdLog, "unknown disk type\n");
        break;
    }
  }
  else
  {
    if ( (*extension == 't') || (*extension == 'T') )
    {
      switch(x)
      {
        case 64:
          rc = IMAGE_TYPE_T64;
          IFDEBUG fprintf(fdLog, "image type = T64\n");
          break;

        default:
          IFDEBUG fprintf(fdLog, "unknown disk type\n");
          break;
      }
    }
    else
      IFDEBUG fprintf(fdLog, "unknown disk type\n");
  }
  return rc;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
HANDLE __stdcall OpenArchive (tOpenArchiveData *ArchiveData)
{
  HANDLE hFile;
  BYTE *pPos;
  DWORD diskSize, bytesRead;
  int rc;
  T_MAIN_STATE *pInfo;
  int diskType=IMAGE_TYPE_NO_TYPE;
  FILE *fdLog=NULL;
  char logFile [MAX_PATH];

  determineIniAndLogFile(&fdLog, logFile);

  // check image extension
  if (strlen(ArchiveData->ArcName) > 4)
  {
    pPos = strrchr(ArchiveData->ArcName, '.');
    if (pPos != NULL)
      diskType = getTypeFromExtension(pPos+1, fdLog);
  }

  // do we have a good type ?
  if (diskType == IMAGE_TYPE_NO_TYPE)
  {
    ArchiveData->OpenResult = E_UNKNOWN_FORMAT;
    IFDEBUG fprintf(fdLog, "unknown disk type\n");
    IFDEBUG fclose(fdLog);
    return NULL;
  }

  IFDEBUG fprintf(fdLog, "Open %s\n", ArchiveData->ArcName);

  hFile = CreateFile(ArchiveData->ArcName, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hFile == INVALID_HANDLE_VALUE)
  {
    IFDEBUG fprintf(fdLog, "error CreateFile %d\n", GetLastError());
    ArchiveData->OpenResult = E_EOPEN;
    IFDEBUG fclose(fdLog);
    return(NULL);
  }

  // get file size
  diskSize = GetFileSize(hFile, NULL);
  if (diskSize == -1)
  {
    IFDEBUG fprintf(fdLog, "error GetFileSize %d\n", GetLastError());
    ArchiveData->OpenResult = E_EREAD;
    IFDEBUG fclose(fdLog);
    return (NULL);
  }

  // allocate disk memory
  pInfo = malloc(sizeof(T_MAIN_STATE));
  if (pInfo == NULL)
  {
    IFDEBUG fprintf(fdLog, "error malloc pInfo\n");
    ArchiveData->OpenResult = E_NO_MEMORY;
    IFDEBUG fclose(fdLog);
    return (NULL);
  }
  memset(pInfo, 0, sizeof(T_MAIN_STATE));

  pInfo->diskData = malloc(diskSize);
  if (pInfo->diskData == NULL)
  {
    IFDEBUG fprintf(fdLog, "error malloc disk data\n");
    ArchiveData->OpenResult = E_NO_MEMORY;
    IFDEBUG fclose(fdLog);
    return (NULL);
  }
  pInfo->diskDataSize = diskSize;
  memset(pInfo->diskData, 0, diskSize);

  // global struct
  pInfo->diskType = diskType;
  pInfo->ArcName = strdup(ArchiveData->ArcName);

  // check image extension
  switch(diskType)
  {
    case IMAGE_TYPE_T64:
      pInfo->state.tapeInfo.currFileSlot = 0;
      pInfo->state.tapeInfo.tapeSize = diskSize;
      IFDEBUG fprintf(fdLog, "image type = T64\n");
      break;

    case IMAGE_TYPE_1541_D64:
      pInfo->state.diskInfo.firstRun = TRUE;
      setNrOfTracks(pInfo);
      pInfo->state.diskInfo.firstDirTrack = 18;
      pInfo->state.diskInfo.firstDirSector = 1;
      pInfo->state.diskInfo.pSectorAdresses = &sectorAdresses_d64[0];
      IFDEBUG fprintf(fdLog, "disk type = D64\n");
      break;

    case IMAGE_TYPE_8050_D80:
    case IMAGE_TYPE_8250_D82:
      pInfo->state.diskInfo.firstRun = TRUE;
      pInfo->state.diskInfo.firstDirTrack = 39;
      pInfo->state.diskInfo.firstDirSector = 1;

      if (diskType == IMAGE_TYPE_8050_D80)
      {
        pInfo->state.diskInfo.nrOfTracks = 77;
        pInfo->state.diskInfo.pSectorAdresses = &sectorAdresses_d80[0];
        IFDEBUG fprintf(fdLog, "disk type = D80\n");
      }
      else
      {
        pInfo->state.diskInfo.nrOfTracks = 154;
        pInfo->state.diskInfo.pSectorAdresses = &sectorAdresses_d82[0];
        IFDEBUG fprintf(fdLog, "disk type = D82\n");
      }
      break;

    case IMAGE_TYPE_NO_TYPE:
      ArchiveData->OpenResult = E_UNKNOWN_FORMAT;
      IFDEBUG fprintf(fdLog, "unknown disk type\n");
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return (NULL);

    default:
      ArchiveData->OpenResult = E_UNKNOWN_FORMAT;
      IFDEBUG fprintf(fdLog, "unknown disk type\n");
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return (NULL);
  }

  // read file at once
  rc = ReadFile(hFile, pInfo->diskData, diskSize, &bytesRead, NULL);
  if (rc == 0 || bytesRead != diskSize)
  {
    IFDEBUG fprintf(fdLog, "error ReadFile %d\n", GetLastError());
    ArchiveData->OpenResult = E_EREAD;
    free (pInfo);
    IFDEBUG fclose(fdLog);
    return (NULL);
  }

  // set tape header pointer
  if (diskType == IMAGE_TYPE_T64)
  {
    pInfo->state.tapeInfo.Header = pInfo->diskData;
    pInfo->state.tapeInfo.Entry  = (TapeEntry *)(pInfo->state.tapeInfo.Header + 1);
    pInfo->state.tapeInfo.Data   = (BYTE *)(pInfo->state.tapeInfo.Entry + pInfo->state.tapeInfo.Header->MaxFiles);

    IFDEBUG fprintf(fdLog, "tape info: header offset = 0x%04X\n", (BYTE *)pInfo->state.tapeInfo.Header-pInfo->diskData);
    IFDEBUG fprintf(fdLog, "tape info: entry offset  = 0x%04X\n", (BYTE *)pInfo->state.tapeInfo.Entry-pInfo->diskData);
    IFDEBUG fprintf(fdLog, "tape info: data offset   = 0x%04X\n", (BYTE *)pInfo->state.tapeInfo.Data-pInfo->diskData);

    IFDEBUG fprintf(fdLog, "tape header: TapeDescr = <%.32s>\n", pInfo->state.tapeInfo.Header->TapeDescr);
    IFDEBUG fprintf(fdLog, "tape header: Version   = <%d.%d>\n", pInfo->state.tapeInfo.Header->Version[1], pInfo->state.tapeInfo.Header->Version[0]);
    IFDEBUG fprintf(fdLog, "tape header: MaxFiles  = <%d>\n", pInfo->state.tapeInfo.Header->MaxFiles);
    IFDEBUG fprintf(fdLog, "tape header: CurrFiles = <%d>\n", pInfo->state.tapeInfo.Header->CurrFiles);
    IFDEBUG fprintf(fdLog, "tape header: UserDescr = <%.24s>\n", pInfo->state.tapeInfo.Header->UserDescr);
  }

  // close file
  CloseHandle(hFile);

  ArchiveData->OpenResult = 0;

  IFDEBUG fclose(fdLog);

  return (pInfo);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
void convertFilename (BYTE *fileName, BYTE fileType)
{
  trimString(fileName);
  // append extension if applicable
  if (fileType >= FILETYPE_DEL && fileType <= FILETYPE_REL)
    strcat(fileName, fileTypes[fileType-FILETYPE_DEL]);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
int __stdcall ReadHeader (HANDLE hArcData, tHeaderData *HeaderData)
{
  T_MAIN_STATE *pInfo=(BYTE *)hArcData;
  int rc, addrDelta, tapeDelta;
  T_DIR_ENTRY *pCurrDirEntry;
  TapeEntry *pCurrTapeEntry;
  BYTE currSector [256];
  FILE *fdLog=NULL;
  char logFile [MAX_PATH];

  determineIniAndLogFile(&fdLog, logFile);

  IFDEBUG fprintf(fdLog, "ReadHeader entry <%s>\n", HeaderData->ArcName);

  if (pInfo->diskType == IMAGE_TYPE_T64)
  {
    // tape images
    memset(HeaderData, 0, sizeof(tHeaderData));

    while (pInfo->state.tapeInfo.currFileSlot < pInfo->state.tapeInfo.Header->MaxFiles)
    {
      pCurrTapeEntry = &pInfo->state.tapeInfo.Entry[pInfo->state.tapeInfo.currFileSlot];
      if (pCurrTapeEntry->EntryUsed == 1)  // normal file, no snapshot
      {
        addrDelta = pCurrTapeEntry->EndAddr-pCurrTapeEntry->StartAddr;
        tapeDelta = pInfo->state.tapeInfo.tapeSize-pCurrTapeEntry->TapePos;

        // correct tape file size if applicable
        if (addrDelta > tapeDelta)
        {
          IFDEBUG fprintf(fdLog, "correcting EndAddr\n");
          pCurrTapeEntry->EndAddr = tapeDelta+pCurrTapeEntry->StartAddr;
        }

        IFDEBUG fprintf(fdLog, "file found in tape slot %d\n", pInfo->state.tapeInfo.currFileSlot);

        convertCBMString(pCurrTapeEntry->FileName, sizeof(pCurrTapeEntry->FileName));

        strcpy(HeaderData->ArcName, pInfo->ArcName);
        strncpy(HeaderData->FileName, pCurrTapeEntry->FileName, sizeof(pCurrTapeEntry->FileName));
        convertFilename(HeaderData->FileName, FILETYPE_PRG);
        HeaderData->PackSize = 0;
        HeaderData->UnpSize  = pCurrTapeEntry->EndAddr-pCurrTapeEntry->StartAddr;
        HeaderData->FileAttr = 0;
        HeaderData->FileTime = ((1980 - 1980) << 25) | (1 << 21) | (1 << 16) | (0 << 11) | (0 << 5) | 0;

        // set next slot to search next time
        pInfo->state.tapeInfo.currFileSlot++;

        // file found
        IFDEBUG fclose (fdLog);
        return 0;
      }
      else
        IFDEBUG fprintf(fdLog, "empty tape slot %d\n", pInfo->state.tapeInfo.currFileSlot);

      // advance search to next slot
      pInfo->state.tapeInfo.currFileSlot++;
    }

    // no more files on tape
    IFDEBUG fprintf(fdLog, "ReadHeader exit E_END_ARCHIVE\n");
    IFDEBUG fclose (fdLog);
    return (E_END_ARCHIVE);
  }
  else
  { // disk images
    while(1)
    {
      memset(HeaderData, 0, sizeof(tHeaderData));

      // determine next block
      if (pInfo->state.diskInfo.firstRun)
      {
        // read first BAM block
        pInfo->state.diskInfo.nextTrack  = pInfo->state.diskInfo.firstDirTrack;
        pInfo->state.diskInfo.nextSector = pInfo->state.diskInfo.firstDirSector;
        pInfo->state.diskInfo.nextEntry  = 0;

        IFDEBUG fprintf(fdLog, "firstSector %d:%d-%d\n", pInfo->state.diskInfo.nextTrack, pInfo->state.diskInfo.nextSector, pInfo->state.diskInfo.nextEntry);

        pInfo->state.diskInfo.firstRun = FALSE;

        // setup cyclic sector access check memory
        memset(pInfo->state.diskInfo.arrCyclicAccessInfo, 0, sizeof(pInfo->state.diskInfo.arrCyclicAccessInfo));
      }
      else
      {
        // not first block
        if (pInfo->state.diskInfo.nextEntry < 7)
          pInfo->state.diskInfo.nextEntry++; // same block, next entry
        else
        {
          // switch block or finish
          if (pInfo->state.diskInfo.currNextTrack != 0)
          {
            pInfo->state.diskInfo.nextTrack  = pInfo->state.diskInfo.currNextTrack;
            pInfo->state.diskInfo.nextSector = pInfo->state.diskInfo.currNextSector;
            pInfo->state.diskInfo.nextEntry = 0;
          }
          else                              // end of BAM chain
          {
            IFDEBUG fprintf(fdLog, "ReadHeader exit E_END_ARCHIVE\n");
            IFDEBUG fclose (fdLog);
            return (E_END_ARCHIVE);
          }
        }
      }

      // read selected sector
      IFDEBUG fprintf(fdLog, "currSector %d:%d-%d\n", pInfo->state.diskInfo.nextTrack, pInfo->state.diskInfo.nextSector, pInfo->state.diskInfo.nextEntry);
      rc = GetDiskSector(pInfo, pInfo->state.diskInfo.nextTrack, pInfo->state.diskInfo.nextSector, currSector);
      if (rc != 0)
      {
        IFDEBUG fprintf(fdLog, "error ReadHeader GetDiskSector %d:%d\n", pInfo->state.diskInfo.nextTrack, pInfo->state.diskInfo.nextSector);
        IFDEBUG fclose (fdLog);
        return (E_EREAD);
      }

      // cyclic check and enter sector index info (but only if first dir entry)
      if (pInfo->state.diskInfo.nextEntry == 0 &&
          pInfo->state.diskInfo.currNextTrack != 0 &&
          pInfo->state.diskInfo.arrCyclicAccessInfo[pInfo->state.diskInfo.nextSector])
      {
        int rc;

        IFDEBUG fprintf(fdLog, "error cyclic access for %d:%d\n", pInfo->state.diskInfo.nextTrack, pInfo->state.diskInfo.nextSector);
        IFDEBUG fclose (fdLog);

        rc = MessageBox(NULL, "Cyclic block chain in directory.\nAborting !!!", "DIRCBM", MB_ICONERROR);

        return (E_EREAD);
      }
      else
        pInfo->state.diskInfo.arrCyclicAccessInfo[pInfo->state.diskInfo.nextSector]++;  // mark sector as read

      // store current next tracks/sector entry
      pInfo->state.diskInfo.currNextTrack  = currSector[0];
      pInfo->state.diskInfo.currNextSector = currSector[1];

      // read next file entry
      pCurrDirEntry = currSector+pInfo->state.diskInfo.nextEntry*sizeof(T_DIR_ENTRY);
      if (pCurrDirEntry->fileType != FILETYPE_SCRATCHED)
      {
        convertCBMString(pCurrDirEntry->filename, sizeof(pCurrDirEntry->filename));

        // store dir entry in master array
        pInfo->state.diskInfo.pFiles = realloc(pInfo->state.diskInfo.pFiles, ++pInfo->state.diskInfo.fileCounter*sizeof(T_DIR_ENTRY));
        pInfo->state.diskInfo.pFiles[pInfo->state.diskInfo.fileCounter-1] = *pCurrDirEntry;

        strcpy(HeaderData->ArcName, pInfo->ArcName);
        strncpy(HeaderData->FileName, pCurrDirEntry->filename, sizeof(pCurrDirEntry->filename));
        convertFilename(HeaderData->FileName, pCurrDirEntry->fileType); // trim blanks and append extension
        HeaderData->PackSize = 0;
        HeaderData->UnpSize  = pCurrDirEntry->nrSectors;
        HeaderData->FileAttr = 0;
        HeaderData->FileTime = ((1980 - 1980) << 25) | (1 << 21) | (1 << 16) | (0 << 11) | (0 << 5) | 0;

        break; // leave loop
      }
      else
        IFDEBUG fprintf(fdLog, "scratched file at %d:%d-%d\n", pInfo->state.diskInfo.nextTrack, pInfo->state.diskInfo.nextSector, pInfo->state.diskInfo.nextEntry);
    }
  }

  // file is read
  IFDEBUG fprintf(fdLog, "ReadHeader exit normal\n");
  IFDEBUG fclose (fdLog);

  return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
TapeEntry *searchTapeEntry(T_MAIN_STATE *pInfo, BYTE *path, BYTE *file, FILE *fdLog)
{
  int i;
  BYTE myFile [30];
  BYTE currFile [30];

  // format file name
  if (path == NULL)
  {
    BYTE *p=strrchr(file, '\\');

    if(p)
      strcpy(myFile, p+1);
  }
  else
    strcpy (myFile, file);

  // later with hashing
  for(i=0; i<pInfo->state.tapeInfo.Header->MaxFiles; i++)
  {
    memset(currFile, 0, sizeof(currFile));
    strncpy(currFile, pInfo->state.tapeInfo.Entry[i].FileName, sizeof(pInfo->state.tapeInfo.Entry[i].FileName));
    convertFilename(currFile, FILETYPE_PRG);
    IFDEBUG fprintf(fdLog, "COMPARE <%s> <%s>\n", myFile, currFile);
    if (strcmp(myFile, currFile) == 0)
      return (&pInfo->state.tapeInfo.Entry[i]);      // found
  }

  return NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
T_DIR_ENTRY *searchDirEntryInDirTrack(T_MAIN_STATE *pInfo, BYTE *filename, FILE *fdLog)
{
  T_DIR_ENTRY *pDirEntry;
  BYTE currDirTrack, currDirSector, *pAddr, x, fileType;
  BOOL rc;
  char buffer [100], *pExt;

  // setup filename and type
  strcpy(buffer, filename);
  pExt = strrchr(buffer, '.');
  if (pExt == NULL)
    return NULL;

  if (stricmp(pExt, ".PRG") == 0)
    fileType = FILETYPE_PRG;
  else
    if (stricmp(pExt, ".SEQ") == 0)
      fileType = FILETYPE_SEQ;
    else
      return NULL;                      // other types are ignored

  // cut off extension
  *pExt = '\0';

  // setup first file track/sector
  currDirTrack = pInfo->state.diskInfo.firstDirTrack;
  currDirSector = pInfo->state.diskInfo.firstDirSector;

  // go through all dir sectors
  while(1)
  {
    rc = GetDiskSectorAddr(pInfo, currDirTrack, currDirSector, &pAddr);
    if (rc != 0)
      return NULL;

    pDirEntry = (T_DIR_ENTRY *)pAddr;

    // search all entries in current dir sector
    for(x=0; x<8; x++)
    {
 IFDEBUG fprintf(fdLog, "searchDirEntryInDirTrack: searching dir=%s search=%s\n", pDirEntry[x].filename, buffer);
      if ( (strncmp(pDirEntry[x].filename, buffer, min(strlen(buffer), sizeof(pDirEntry[x].filename))) == 0) &&
           (fileType == pDirEntry[x].fileType) )
      {
        IFDEBUG fprintf(fdLog, "searchDirEntryInDirTrack: file found at %d:%d-%d\n", currDirTrack, currDirSector, x);
        return &pDirEntry[x];
      }
    }

    // switch to next dir sector
    currDirTrack = *pAddr;
    currDirSector = *(pAddr+1);
  }

  return NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
T_DIR_ENTRY *searchDirEntry(T_MAIN_STATE *pInfo, BYTE *path, BYTE *file, FILE *fdLog)
{
  int i;
  BYTE myFile [30];
  BYTE currFile [30];

  // format file name
  if (path == NULL)
  {
    BYTE *p=strrchr(file, '\\');

    if(p)
      strcpy(myFile, p+1);
  }
  else
    strcpy (myFile, file);

  // later with hashing
  IFDEBUG fprintf(fdLog, "COMPARE fileCounter %d\n", pInfo->state.diskInfo.fileCounter);
  for(i=0; i<pInfo->state.diskInfo.fileCounter; i++)
  {
    memset(currFile, 0, sizeof(currFile));
    strncpy(currFile, pInfo->state.diskInfo.pFiles[i].filename, sizeof(pInfo->state.diskInfo.pFiles[i].filename));
    convertFilename(currFile, pInfo->state.diskInfo.pFiles[i].fileType);
    IFDEBUG fprintf(fdLog, "COMPARE <%s> <%s>\n", myFile, currFile);
    if (strcmp(myFile, currFile) == 0)
      return (&pInfo->state.diskInfo.pFiles[i]);      // found
  }

  return NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
int __stdcall ProcessFile (HANDLE hArcData, int Operation, char *DestPath, char *DestName)
{
  T_MAIN_STATE *pInfo=(BYTE *)hArcData;
  T_DIR_ENTRY *pFile;
  TapeEntry *pTapeFile;
  FILE *fdLog=NULL;
  char logFile [MAX_PATH];

  // test order
  if (Operation == PK_SKIP)
    return 0;

  determineIniAndLogFile(&fdLog, logFile);

  IFDEBUG fprintf(fdLog, "ProcessFile - ");
  switch (Operation)
  {
    case PK_SKIP:
      // IFDEBUG fprintf(fdLog, "SKIP <%s> <%s>\n", DestPath, DestName);
      break;

    case PK_EXTRACT:  // extract c64 file to temp file
// ProcessFile - EXTRACT <(null)> <C:\DOCUME~1\atw10qy4\LOCALS~1\Temp\_tc\copy vers.0.6.prg>
      IFDEBUG fprintf(fdLog, "ProcessFile - EXTRACT <%s> <%s>\n", DestPath, DestName);

      if (pInfo->diskType == IMAGE_TYPE_T64)
      { // search tape
        if (pTapeFile = searchTapeEntry(pInfo, DestPath, DestName, fdLog))
        {
          FILE *fd;

          IFDEBUG fprintf(fdLog, "found: TapePos   = %d\n", pTapeFile->TapePos);
          IFDEBUG fprintf(fdLog, "found: StartAddr = %d\n", pTapeFile->StartAddr);
          IFDEBUG fprintf(fdLog, "found: EndAddr   = %d\n", pTapeFile->EndAddr);

          fd = fopen (DestName, "wb");
          if (fd == NULL)
          {
            IFDEBUG fprintf(fdLog, "error fopen <%s>\n", DestName);
            IFDEBUG fclose(fdLog);
            return (E_EWRITE);
          }
          fwrite(&pTapeFile->StartAddr, sizeof(WORD), 1, fd);
          fwrite(pInfo->diskData+pTapeFile->TapePos, pTapeFile->EndAddr-pTapeFile->StartAddr, 1, fd);
          fclose(fd);

          IFDEBUG fclose(fdLog);

          return 0;                     // success
        }
        else
        {
          IFDEBUG fclose(fdLog);
          return E_NO_FILES;
        }
      }
      else
      { // search disk
        int arrCyclicAccessInfo[MAX_OVERALL_SECTORS];

        // setup buffer
        memset(arrCyclicAccessInfo, 0, sizeof(arrCyclicAccessInfo));

        if (pFile = searchDirEntry(pInfo, DestPath, DestName, fdLog))
        {
          FILE *fd;
          BYTE currSector [256];
          int rc;
          BYTE nextTrack, nextSector;
          int offset;

          IFDEBUG fprintf(fdLog, "EXTRACT found <%s> %d\n", pFile->filename, pFile->nrSectors);

          fd = fopen (DestName, "wb");
          if (fd == NULL)
          {
            IFDEBUG fprintf(fdLog, "error fopen <%s>\n", DestName);
            IFDEBUG fclose(fdLog);
            return (E_EWRITE);
          }

          nextTrack  = pFile->firstFileTrack;
          nextSector = pFile->firstFileSector;
          while(1)
          {
            rc = GetDiskSector(pInfo, nextTrack, nextSector, currSector);
            if (rc != 0)
            {
              IFDEBUG fprintf(fdLog, "error Extract GetDiskSector %d:%d\n", nextTrack, nextSector);
              IFDEBUG fclose(fdLog);
              return (E_EREAD);
            }

            // check if cyclic sector chain
            offset = nextTrack*MAX_SECTORS_PER_TRACK+nextSector;
            if (arrCyclicAccessInfo[offset])
            {
              int rc;
              char buffer [200];

              IFDEBUG fprintf(fdLog, "error cyclic access for %d:%d\n", nextTrack, nextSector);
              IFDEBUG fclose (fdLog);

              sprintf(buffer, "Cyclic block chain in file \"%s\".\nAborting !!!", pFile->filename);
              rc = MessageBox(NULL, buffer, "DIRCBM", MB_ICONERROR);

              fclose(fd);

              return (E_EREAD);
            }
            else
              arrCyclicAccessInfo[offset]++;  // mark sector as read

            // last sector ??
            if (currSector[0] == 0)
            {
              fwrite(currSector+2, currSector[1]-2+1, 1, fd);
              fclose(fd);
              break;
            }

            fwrite(currSector+2, 254, 1, fd);
            nextTrack  = currSector[0];
            nextSector = currSector[1];
          }
        }
        else
        {
          IFDEBUG fclose(fdLog);
          return E_NO_FILES;
        }
      }
      break;

    case PK_TEST:
      IFDEBUG fprintf(fdLog, "ProcessFile - TEST <%s> <%s>\n", DestPath, DestName);
      break;

    default:
      IFDEBUG fprintf(fdLog, "ProcessFile - other operation %d <%s> <%s>\n", Operation, DestPath, DestName);
      break;
  }

  IFDEBUG fclose(fdLog);

  return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
int __stdcall CloseArchive (HANDLE hArcData)
{
  T_MAIN_STATE *pInfo=(T_MAIN_STATE *)((BYTE *)hArcData);
  FILE *fdLog=NULL;
  char logFile [MAX_PATH];

  determineIniAndLogFile(&fdLog, logFile);

  IFDEBUG fprintf(fdLog, "CloseArchive %04X\n", pInfo);
  IFDEBUG fclose(fdLog);

  free(pInfo->ArcName); // archive name
  free(pInfo->state.diskInfo.pFiles);  // dir entry list
  free(pInfo);  // free allocated memory from OpenArchive (main info + disk image buffer)
  IFDEBUG fclose(fdLog);

  return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
int __stdcall GetPackerCaps()
{
  FILE *fdLog=NULL;
  char logFile [MAX_PATH];

  determineIniAndLogFile(&fdLog, logFile);
  IFDEBUG fprintf(fdLog, "GetPackerCaps\n");
  IFDEBUG fclose(fdLog);

  return PK_CAPS_NEW | PK_CAPS_MODIFY | PK_CAPS_MULTIPLE | PK_CAPS_DELETE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
void fillFirstD64BAMBlock (T_MAIN_STATE *pInfo, FILE *fdLog)
{
  DWORD value;
  int x, rc;
  BYTE *pAddr;
  T_BAM_1541 *pBlock;

  rc = GetDiskSectorAddr(pInfo, pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector, &pAddr);
  if(rc != 0)
  {
    IFDEBUG fprintf(fdLog, "fillFirstD64BAMBlock: bad GetDiskSectorAddr call (%d:%d)\n", pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector);
    return;
  }
  pBlock = (T_BAM_1541 *)pAddr;

  memset(pAddr, 0, 256+2);

  pBlock->nextDirTrack = pInfo->state.diskInfo.firstBAMTrack;
  pBlock->nextDirSector = 1;
  pBlock->formatID = 'A';

  for (x=1; x<=pInfo->state.diskInfo.nrOfTracks; x++)
  {
    if (x != pInfo->state.diskInfo.firstBAMTrack)
    {
      // setup data track info
      pBlock->BAM_Data[x-1].sectorsLeft = pInfo->state.diskInfo.pSectorAdresses[x-1][BLOCK_INFO];
      value = (1 << pBlock->BAM_Data[x-1].sectorsLeft) - 1;
    }
    else
    {
      // setup BAM track info
      // zwei Blcke bereits verbraucht
      pBlock->BAM_Data[x-1].sectorsLeft = pInfo->state.diskInfo.pSectorAdresses[x-1][BLOCK_INFO];
      value = ((1 << pBlock->BAM_Data[x-1].sectorsLeft) - 1) & 0x00fffffc;
      pBlock->BAM_Data[x-1].sectorsLeft -= 2;
    }
    memcpy(pBlock->BAM_Data[x-1].BAMOfTrack, &value, 3);
  }
  memset(pBlock->diskName, 0xA0, sizeof(pBlock->diskName));
  strncpy(pBlock->diskName, "DIRCBM", 6);
  strncpy(pBlock->diskName+7, "DISK", 4);
  strncpy(pBlock->diskID, "00", 2);
  pBlock->fillByte1 = 0xA0;
  strncpy(pBlock->formatIDName, "2A", 2);
  *((UINT32 *)pBlock->fillBytes) = 0xA0A0A0A0;

  *(pAddr+257) = (BYTE)0xff;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
int getFreeDataBlocksOnDisk(T_MAIN_STATE *pInfo, FILE *fdLog)
{
  int blocksFree=0, x, rc;
  T_BAM_1541 *pBlock;
  BYTE *pAddr;

  rc = GetDiskSectorAddr(pInfo, pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector, &pAddr);
  if(rc != 0)
  {
    IFDEBUG fprintf(fdLog, "getFreeDataBlocksOnDisk: bad GetDiskSectorAddr call (%d:%d)\n", pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector);
    return 0;
  }
  pBlock = (T_BAM_1541 *)pAddr;

  for (x=1; x<=pInfo->state.diskInfo.nrOfTracks; x++)
    if (x != pInfo->state.diskInfo.firstBAMTrack)
      blocksFree += pBlock->BAM_Data[x-1].sectorsLeft;

  return blocksFree;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
int getFreeDirBlocksOnDisk(T_MAIN_STATE *pInfo, FILE *fdLog)
{
  int blocksFree=0, x, rc;
  T_BAM_1541 *pBlock;
  BYTE *pAddr;

  rc = GetDiskSectorAddr(pInfo, pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector, &pAddr);
  if(rc != 0)
  {
    IFDEBUG fprintf(fdLog, "getFreeDirBlocksOnDisk: bad GetDiskSectorAddr call (%d:%d)\n", pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector);
    return 0;
  }
  pBlock = (T_BAM_1541 *)pAddr;

  return pBlock->BAM_Data[pInfo->state.diskInfo.firstBAMTrack-1].sectorsLeft;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
BYTE getFreeSectorFromBAM(T_MAIN_STATE *pInfo, BYTE track, FILE *fdLog)
{
  BYTE maxBlocksInTrack=pInfo->state.diskInfo.pSectorAdresses[track-1][BLOCK_INFO];
  T_BAM_1541 *pBlock;
  BYTE *pAddr, i;
  int value, rc;

  rc = GetDiskSectorAddr(pInfo, pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector, &pAddr);
  if(rc != 0)
  {
    IFDEBUG fprintf(fdLog, "getFreeSectorFromBAM: bad GetDiskSectorAddr call (%d:%d)\n", pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector);
    return 0xff;
  }
  pBlock = (T_BAM_1541 *)pAddr;

  if (pBlock->BAM_Data[track-1].sectorsLeft == 0)
    return 0xff;

  value = *((int *)pBlock->BAM_Data[track-1].BAMOfTrack) & 0x00FFFFFF;
  i = 0;
  while( (((1 << i) & value) == 0) && (i < maxBlocksInTrack) )
    i++;

  if (i == maxBlocksInTrack)
    return 0xff; // no blocks left
  else
    return i;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
T_DIR_ENTRY *getFreeDirEntry(T_MAIN_STATE *pInfo, FILE *fdLog)
{
  int x, y, blocksFree, rc;
  BYTE *pAddr, nextDirTrack, nextDirSector, currDirTrack, currDirSector, lastDirTrack, lastDirSector, newSector;
  T_DIR_ENTRY *pDirEntry;

  // get free blocks in dir track
  blocksFree = getFreeDirBlocksOnDisk(pInfo, fdLog);
  IFDEBUG fprintf(fdLog, "getFreeDirEntry: %d dir blocks available\n", blocksFree);

  currDirTrack  = pInfo->state.diskInfo.firstDirTrack;
  currDirSector = pInfo->state.diskInfo.firstDirSector;

  // search free dir entry
  rc = GetDiskSectorAddr(pInfo, currDirTrack, currDirSector, &pAddr);
  if(rc != 0)
  {
    IFDEBUG fprintf(fdLog, "getFreeDirEntry: bad GetDiskSectorAddr call (%d:%d)\n", currDirTrack, currDirSector);
    return NULL;
  }
  pDirEntry = (T_DIR_ENTRY *)pAddr;

  while(1)
  {
    IFDEBUG fprintf(fdLog, "getFreeDirEntry: searching %d:%d ...\n", currDirTrack, currDirSector);

    nextDirTrack = pDirEntry->nextDirTrack;
    nextDirSector = pDirEntry->nextDirSector;

    for(x=0; x<8; x++, pDirEntry++)
      if (pDirEntry->fileType == 0)       // entry free ?
        break;

    if (x < 8)
    {
      IFDEBUG fprintf(fdLog, "getFreeDirEntry: found free entry at %d:%d index %d\n", currDirTrack, currDirSector, x);
      break; // free entry in this block -> leave
    }
    else
    {
      // store last dir track/sector
      lastDirTrack = currDirTrack;
      lastDirSector = currDirSector;

      if (nextDirTrack)                 // further blocks available
      {
        // switch to next block
        rc = GetDiskSectorAddr(pInfo, nextDirTrack, nextDirSector, &pAddr);
        if(rc != 0)
        {
          IFDEBUG fprintf(fdLog, "getFreeDirEntry: bad GetDiskSectorAddr call (%d:%d)\n", nextDirTrack, nextDirSector);
          return NULL;
        }
        pDirEntry = (T_DIR_ENTRY *)pAddr;

        currDirTrack  = nextDirTrack;
        currDirSector = nextDirSector;
      }
      else
      {
        // no free entry in this block -> create new one if available
        if (blocksFree == 0)
          return NULL;
        else
        {
          // allocate new dir block
          newSector = getFreeSectorFromBAM(pInfo, currDirTrack, fdLog);
          if (newSector == 0xff)
            return NULL;
          markSectorAsUsed(pInfo, currDirTrack, newSector, fdLog);

          // we have a new dir sector
          rc = GetDiskSectorAddr(pInfo, currDirTrack, newSector, &pAddr);
          if(rc != 0)
          {
            IFDEBUG fprintf(fdLog, "getFreeDirEntry: bad GetDiskSectorAddr call (%d:%d)\n", currDirTrack, newSector);
            return NULL;
          }
          memset(pAddr, 0, 256);
          *(pAddr+1) = 0xff;
          pDirEntry = (T_DIR_ENTRY *)pAddr;

          // update link info from last sector
          rc = GetDiskSectorAddr(pInfo, lastDirTrack, lastDirSector, &pAddr);
          if(rc != 0)
          {
            IFDEBUG fprintf(fdLog, "getFreeDirEntry: bad GetDiskSectorAddr call (%d:%d)\n", lastDirTrack, lastDirSector);
            return NULL;
          }
          *pAddr = currDirTrack;
          *(pAddr+1) = newSector;

          currDirTrack  = currDirTrack;
          currDirSector = newSector;

          IFDEBUG fprintf(fdLog, "getFreeDirEntry: new entry at %d:%d\n", currDirTrack, newSector);
        }
      }
    }
  }

  // pDirEntry is pointer to current entry
  return pDirEntry;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
BYTE *getFreeDataBlock(T_MAIN_STATE *pInfo, BYTE *currTrack, BYTE *currSector, FILE *fdLog)
{
  int rc;
  BYTE *pAddr, newSector, x;

  for (x=1; x<=pInfo->state.diskInfo.nrOfTracks; x++)
  {
    if (x != pInfo->state.diskInfo.firstDirTrack)
    {
      newSector = getFreeSectorFromBAM(pInfo, x, fdLog);
      if (newSector != 0xff)
      {
        *currTrack = x;
        *currSector = newSector;
        rc = GetDiskSectorAddr(pInfo, *currTrack, *currSector, &pAddr);
        if(rc != 0)
        {
          IFDEBUG fprintf(fdLog, "getFreeDataBlock: bad GetDiskSectorAddr call (%d:%d)\n", *currTrack, *currSector);
          return NULL;
        }
        return pAddr;
      }
    }
  }
  return NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
void markSectorAsFree(T_MAIN_STATE *pInfo, BYTE currTrack, BYTE currSector, FILE *fdLog)
{
  T_BAM_1541 *pBlock;
  BYTE *pAddr;
  int value, rc;

  // check existence of data sector
  rc = GetDiskSectorAddr(pInfo, currTrack, currSector, &pAddr);
  if(rc != 0)
  {
    IFDEBUG fprintf(fdLog, "markSectorAsFree: bad GetDiskSectorAddr call data (%d:%d)\n", currTrack, currSector);
    return;
  }

  // get BAM sector
  rc = GetDiskSectorAddr(pInfo, pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector, &pAddr);
  if(rc != 0)
  {
    IFDEBUG fprintf(fdLog, "markSectorAsFree: bad GetDiskSectorAddr call BAM (%d:%d)\n", pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector);
    return;
  }
  pBlock = (T_BAM_1541 *)pAddr;

  IFDEBUG fprintf(fdLog, "markSectorAsFree: BAMtrack=%d, BAMsector=%d\n", pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector);

  value = *((int *)pBlock->BAM_Data[currTrack-1].BAMOfTrack) & 0x00FFFFFF;
  pBlock->BAM_Data[currTrack-1].sectorsLeft++;

  IFDEBUG fprintf(fdLog, "markSectorAsFree: value=%x, track=%d, sector=%d\n", value, currTrack, currSector);

  value |= 1 << currSector;
  value &= 0x00FFFFFF;

  memcpy(pBlock->BAM_Data[currTrack-1].BAMOfTrack, &value, 3);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
void markSectorAsUsed(T_MAIN_STATE *pInfo, BYTE currTrack, BYTE currSector, FILE *fdLog)
{
  T_BAM_1541 *pBlock;
  BYTE *pAddr;
  int value, rc;

  // check existence of data sector
  rc = GetDiskSectorAddr(pInfo, currTrack, currSector, &pAddr);
  if(rc != 0)
  {
    IFDEBUG fprintf(fdLog, "markSectorAsUsed: bad GetDiskSectorAddr call data (%d:%d)\n", currTrack, currSector);
    return;
  }

  // get BAM sector
  rc = GetDiskSectorAddr(pInfo, pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector, &pAddr);
  if(rc != 0)
  {
    IFDEBUG fprintf(fdLog, "markSectorAsUsed: bad GetDiskSectorAddr call (%d:%d)\n", pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector);
    return;
  }
  pBlock = (T_BAM_1541 *)pAddr;

  IFDEBUG fprintf(fdLog, "markSectorAsUsed: BAMtrack=%d, BAMsector=%d\n", pInfo->state.diskInfo.firstBAMTrack, pInfo->state.diskInfo.firstBAMSector);

  value = *((int *)pBlock->BAM_Data[currTrack-1].BAMOfTrack) & 0x00FFFFFF;
  pBlock->BAM_Data[currTrack-1].sectorsLeft--;

  IFDEBUG fprintf(fdLog, "markSectorAsUsed: value=%x, track=%d, sector=%d\n", value, currTrack, currSector);

  value &= ~(1 << currSector);
  value &= 0x00FFFFFF;

  memcpy(pBlock->BAM_Data[currTrack-1].BAMOfTrack, &value, 3);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
int __stdcall PackFiles(char *PackedFile, char *SubPath, char *SrcPath, char *AddList, int Flags)
{
  FILE *fdArchive;
  char *pCurrFile;
  BOOL archiveWasExisting;
  BYTE *pImage;
  int diskType=IMAGE_TYPE_NO_TYPE, rc, bytesRead;
  WIN32_FIND_DATA fileInfo;
  HANDLE hFile;
  T_MAIN_STATE *pInfo;
  BYTE *pBAMSector, *pPos;
  FILE *fdLog=NULL;
  char logFile [MAX_PATH];

  determineIniAndLogFile(&fdLog, logFile);

  IFDEBUG fprintf(fdLog, "PackFiles: PackedFile=%s, SubPath=%s, SrcPath=%s\n", PackedFile, SubPath, SrcPath);

  // get disk type from name
  pPos = strrchr(PackedFile, '.');
  if (pPos != NULL)
    diskType = getTypeFromExtension(pPos+1, fdLog);

  // only disk images are supported now
  if (diskType == IMAGE_TYPE_T64 || diskType == IMAGE_TYPE_NO_TYPE)
  {
    IFDEBUG fclose(fdLog);
    return E_NOT_SUPPORTED;
  }

  pInfo = malloc(sizeof(T_MAIN_STATE));
  pInfo->ArcName = strdup(PackedFile);
  pInfo->diskType = diskType;

  // get info if file is exisiting and size if available
  hFile = FindFirstFile(PackedFile, &fileInfo);
  if (hFile != INVALID_HANDLE_VALUE)
  {
    // file is available
    FindClose(hFile);
    IFDEBUG fprintf(fdLog, "opening existing pack file <%s>\n", PackedFile);
    IFDEBUG fprintf(fdLog, "%d, %d\n", fileInfo.nFileSizeHigh, fileInfo.nFileSizeLow);

    pInfo->diskData = malloc(fileInfo.nFileSizeLow);
    if (pInfo->diskData == NULL)
    {
      IFDEBUG fprintf(fdLog, "error malloc disk data\n");
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return E_NO_MEMORY;
    }
    pInfo->diskDataSize = fileInfo.nFileSizeLow;
    memset(pInfo->diskData, 0, fileInfo.nFileSizeLow);

    // read existing disk images
    hFile = CreateFile(PackedFile, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {
      IFDEBUG fprintf(fdLog, "error CreateFile %d\n", GetLastError());
      free(pInfo->diskData);
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return E_EOPEN;
    }

    rc = ReadFile(hFile, pInfo->diskData, fileInfo.nFileSizeLow, &bytesRead, NULL);
    if (rc == 0 || bytesRead != fileInfo.nFileSizeLow)
    {
      IFDEBUG fprintf(fdLog, "error ReadFile %d\n", GetLastError());
      free(pInfo->diskData);
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return E_EREAD;
    }

    rc = CloseHandle(hFile);

    //
    switch(diskType)
    {
      case IMAGE_TYPE_1541_D64:
        // create new image with DIRCBM title like loo.d64 (empty)
        // image is 0-filled, block 18/00 with info, block 18/01 has only 00 FF
        pInfo->state.diskInfo.nrOfTracks = 35;
        pInfo->state.diskInfo.pSectorAdresses = &sectorAdresses_d64[0];
        pInfo->state.diskInfo.firstBAMTrack = 18;
        pInfo->state.diskInfo.firstBAMSector = 0;
        pInfo->state.diskInfo.firstDirTrack = 18;
        pInfo->state.diskInfo.firstDirSector = 1;

        // fill first BAM block
   //     fillFirstD64BAMBlock (pInfo);
        break;

      case IMAGE_TYPE_8050_D80:
        free(pInfo->diskData);
        free(pInfo);
        IFDEBUG fclose(fdLog);
        return E_NOT_SUPPORTED;
        break;

      case IMAGE_TYPE_8250_D82:
        free(pInfo->diskData);
        free(pInfo);
        IFDEBUG fclose(fdLog);
        return E_NOT_SUPPORTED;
        break;

      default:
        free(pInfo->diskData);
        free(pInfo);
        IFDEBUG fclose(fdLog);
        return E_NOT_SUPPORTED;
        break;
    }
  }
  else
  {
    // archive not available --> create new archive (only for dxx formats)
    IFDEBUG fprintf(fdLog, "create new pack file <%s>\n", PackedFile);

    // create empty disk images
    switch(diskType)
    {
      case IMAGE_TYPE_1541_D64:
        // create new image with DIRCBM title like loo.d64 (empty)
        // image is 0-filled, block 18/00 with info, block 18/01 has only 00 FF
        pInfo->diskData = malloc(174848);        // std 1541 disk
        pInfo->diskDataSize = 174848;
        memset(pInfo->diskData, 0, pInfo->diskDataSize);
        pInfo->state.diskInfo.nrOfTracks = 35;
        pInfo->state.diskInfo.pSectorAdresses = &sectorAdresses_d64[0];
        pInfo->state.diskInfo.firstBAMTrack = 18;
        pInfo->state.diskInfo.firstBAMSector = 0;
        pInfo->state.diskInfo.firstDirTrack = 18;
        pInfo->state.diskInfo.firstDirSector = 1;

        // fill first BAM block
        fillFirstD64BAMBlock (pInfo, fdLog);
        break;

      case IMAGE_TYPE_8050_D80:
        free(pInfo->diskData);
        free(pInfo);
        IFDEBUG fclose(fdLog);
        return E_NOT_SUPPORTED;
        break;

      case IMAGE_TYPE_8250_D82:
        free(pInfo->diskData);
        free(pInfo);
        IFDEBUG fclose(fdLog);
        return E_NOT_SUPPORTED;
        break;

      default:
        free(pInfo->diskData);
        free(pInfo);
        IFDEBUG fclose(fdLog);
        return E_NOT_SUPPORTED;
        break;
    }
  }

  // archive is ready to update in memory

  // archive available --> add files to archive
  pCurrFile = AddList;
  while(*pCurrFile)
  {
    char nameBuffer [_MAX_PATH];
    int blocksToWrite, blocksFree, x;
    FILE *fd;
    BYTE *pPrevData=NULL, *pData, currTrack, currSector, *pExt;
    T_DIR_ENTRY *pDirEntry=NULL;
    char *newFilename;

    strcpy(nameBuffer, SrcPath);
    strcat(nameBuffer, pCurrFile);

    IFDEBUG fprintf(fdLog, "adding file <%s>\n", nameBuffer);

    // get info if file is existing and size if available
    hFile = FindFirstFile(nameBuffer, &fileInfo);
    if (hFile == INVALID_HANDLE_VALUE)
    {
      IFDEBUG fprintf(fdLog, "error getting file size for <%s>\n", nameBuffer);
      free(pInfo->diskData);
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return E_EREAD;
    }
    FindClose(hFile);

    // check disk image space
    blocksToWrite = fileInfo.nFileSizeLow / 254 + ((fileInfo.nFileSizeLow % 254) ? 1 : 0 );
    blocksFree = getFreeDataBlocksOnDisk(pInfo, fdLog);
    if (blocksToWrite > blocksFree)
    {
      IFDEBUG fprintf(fdLog, "error not enough space <%d>\n", blocksToWrite);
      free(pInfo->diskData);
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return E_NO_MEMORY;
    }
    IFDEBUG fprintf(fdLog, "blocks to write <%d>, blocks free <%d>\n", blocksToWrite, blocksFree);

    // get dir entry
    pDirEntry = getFreeDirEntry(pInfo, fdLog);
    if (pDirEntry == NULL)
    {
      IFDEBUG fprintf(fdLog, "error: no free dir entry\n");
      free(pInfo->diskData);
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return E_TOO_MANY_FILES;
    }

    // init dir info
    memset(((BYTE *)pDirEntry)+2, 0, sizeof(T_DIR_ENTRY)-2);
    pDirEntry->fileType = FILETYPE_PRG;

    // setup file name buffer and init
    newFilename = strdup(pCurrFile);
    convertPCString(newFilename, strlen(newFilename));
    memset(pDirEntry->filename, 0xA0, sizeof(pDirEntry->filename));

    // do we have an known extension ?
    pExt = strrchr(newFilename, '.');
    if (pExt)
    {
      if (stricmp(pExt, ".PRG") == 0)
      {
        *pExt = '\0';
        pDirEntry->fileType = FILETYPE_PRG;
      }
      else
      {
        if (stricmp(pExt, ".SEQ") == 0)
        {
          *pExt = '\0';
          pDirEntry->fileType = FILETYPE_SEQ;
        }
        else
          pDirEntry->fileType = FILETYPE_SEQ;
      }
    }
    else
      pDirEntry->fileType = FILETYPE_SEQ;  // set SEQ file type for all files other than .PRG

    // copy filename
    strncpy(pDirEntry->filename, newFilename, min(strlen(newFilename), sizeof(pDirEntry->filename)));
    free(newFilename);

    pDirEntry->nrSectors = blocksToWrite;

    // write data to image
    fd = fopen(nameBuffer, "rb");
    if (fd == NULL)
    {
      IFDEBUG fprintf(fdLog, "error reading file <%s>\n", nameBuffer);
      free(pInfo->diskData);
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return E_EREAD;
    }

    // write block loop
    for (x=0; x<blocksToWrite; x++)
    {
      pData = getFreeDataBlock(pInfo, &currTrack, &currSector, fdLog);  // get free block and mark as not free
      if (pData == NULL)
      {
        IFDEBUG fprintf(fdLog, "error: no free blocks\n");
        free(pInfo->diskData);
        free(pInfo);
        IFDEBUG fclose(fdLog);
        return E_NO_MEMORY;
      }

      IFDEBUG fprintf(fdLog, "writing sector %d:%d\n", currTrack, currSector);

      if (x == 0)
      {
        // store first track and sector of file chain
        pDirEntry->firstFileTrack = currTrack;
        pDirEntry->firstFileSector = currSector;
      }

      // update last block link info
      if (pPrevData)
      {
        *pPrevData = currTrack;
        *(pPrevData+1) = currSector;
      }

      memset(pData, 0, 256);
      rc = fread(pData+2, 1, 254, fd);

      IFDEBUG fprintf(fdLog, "marking sector %d:%d\n", currTrack, currSector);
      markSectorAsUsed(pInfo, currTrack, currSector, fdLog);

      pPrevData = pData;
    }

    // write last block header info
    *pPrevData = 0;
    *(pPrevData+1) = rc+2-1;

    fclose(fd);

    // next file
    pCurrFile += strlen(pCurrFile) + 1;
  }

  IFDEBUG fprintf(fdLog, "packing done\n");

  // write image in memory to disk
  fdArchive = fopen(PackedFile, "wb");
  if (fdArchive == NULL)
  {
    IFDEBUG fprintf(fdLog, "error: fopen (write image) failed\n");
    free(pInfo->diskData);
    free(pInfo);
    IFDEBUG fclose(fdLog);
    return E_EWRITE;
  }
  rc = fwrite(pInfo->diskData, 256, 683, fdArchive);  // TODO: !!! 1541 specific
  if (rc != 683)                        // TODO: !!! 1541 specific
  {
    IFDEBUG fprintf(fdLog, "error: fwrite (write image) wrote less (%d != 683)\n", rc);
    free(pInfo->diskData);
    free(pInfo);
    IFDEBUG fclose(fdLog);
    return E_EWRITE;
  }
  fclose(fdArchive);

  free(pInfo->diskData);
  free(pInfo);

  IFDEBUG fclose(fdLog);

  return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
int __stdcall DeleteFiles(char *PackedFile, char *DeleteList)
{
  FILE *fdArchive;
  char *pCurrFile;
  BOOL archiveWasExisting;
  BYTE *pImage;
  int diskType=IMAGE_TYPE_NO_TYPE, rc, bytesRead;
  WIN32_FIND_DATA fileInfo;
  HANDLE hFile;
  T_MAIN_STATE *pInfo;
  BYTE *pBAMSector, *pPos;
  FILE *fdLog=NULL;
  char logFile [MAX_PATH];
  BOOL imageChanged=FALSE;

  determineIniAndLogFile(&fdLog, logFile);

  IFDEBUG fprintf(fdLog, "DeleteFiles: PackedFile=%s\n", PackedFile);

  // get disk type from name
  pPos = strrchr(PackedFile, '.');
  if (pPos != NULL)
    diskType = getTypeFromExtension(pPos+1, fdLog);

  // only disk images are supported now
  if (diskType == IMAGE_TYPE_T64 || diskType == IMAGE_TYPE_NO_TYPE)
  {
    IFDEBUG fclose(fdLog);
    return E_NOT_SUPPORTED;
  }

  pInfo = malloc(sizeof(T_MAIN_STATE));
  pInfo->ArcName = strdup(PackedFile);
  pInfo->diskType = diskType;

  // get info if file is exisiting and size if available
  hFile = FindFirstFile(PackedFile, &fileInfo);
  if (hFile != INVALID_HANDLE_VALUE)
  {
    // file is available
    FindClose(hFile);
    IFDEBUG fprintf(fdLog, "opening existing pack file <%s>\n", PackedFile);
    IFDEBUG fprintf(fdLog, "%d, %d\n", fileInfo.nFileSizeHigh, fileInfo.nFileSizeLow);

    pInfo->diskData = malloc(fileInfo.nFileSizeLow);
    if (pInfo->diskData == NULL)
    {
      IFDEBUG fprintf(fdLog, "error malloc disk data\n");
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return E_NO_MEMORY;
    }
    pInfo->diskDataSize = fileInfo.nFileSizeLow;
    memset(pInfo->diskData, 0, fileInfo.nFileSizeLow);

    // read existing disk images
    hFile = CreateFile(PackedFile, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {
      IFDEBUG fprintf(fdLog, "error CreateFile %d\n", GetLastError());
      free(pInfo->diskData);
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return E_EOPEN;
    }

    rc = ReadFile(hFile, pInfo->diskData, fileInfo.nFileSizeLow, &bytesRead, NULL);
    if (rc == 0 || bytesRead != fileInfo.nFileSizeLow)
    {
      IFDEBUG fprintf(fdLog, "error ReadFile %d\n", GetLastError());
      free(pInfo->diskData);
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return E_EREAD;
    }

    rc = CloseHandle(hFile);

    //
    switch(diskType)
    {
      case IMAGE_TYPE_1541_D64:
        // create new image with DIRCBM title like loo.d64 (empty)
        // image is 0-filled, block 18/00 with info, block 18/01 has only 00 FF
        pInfo->state.diskInfo.nrOfTracks = 35;
        pInfo->state.diskInfo.pSectorAdresses = &sectorAdresses_d64[0];
        pInfo->state.diskInfo.firstBAMTrack = 18;
        pInfo->state.diskInfo.firstBAMSector = 0;
        pInfo->state.diskInfo.firstDirTrack = 18;
        pInfo->state.diskInfo.firstDirSector = 1;

        // fill first BAM block
   //     fillFirstD64BAMBlock (pInfo);
        break;

      case IMAGE_TYPE_8050_D80:
        free(pInfo->diskData);
        free(pInfo);
        IFDEBUG fclose(fdLog);
        return E_NOT_SUPPORTED;
        break;

      case IMAGE_TYPE_8250_D82:
        free(pInfo->diskData);
        free(pInfo);
        IFDEBUG fclose(fdLog);
        return E_NOT_SUPPORTED;
        break;

      default:
        free(pInfo->diskData);
        free(pInfo);
        IFDEBUG fclose(fdLog);
        return E_NOT_SUPPORTED;
        break;
    }
  }
  else
  {
    IFDEBUG fprintf(fdLog, "error CreateFile %d\n", GetLastError());
    free(pInfo->diskData);
    free(pInfo);
    IFDEBUG fclose(fdLog);
    return E_EOPEN;
  }

  // archive is ready to update in memory

  // archive available --> delete files from archive
  pCurrFile = DeleteList;
  while(*pCurrFile)
  {
    int blocksToWrite, blocksFree, x;
    FILE *fd;
    BYTE *pPrevData=NULL, *pData, nextTrack, nextSector, *pAddr;
    T_DIR_ENTRY *pDirEntry=NULL;
    char *newFilename;

    IFDEBUG fprintf(fdLog, "deleting file <%s>\n", pCurrFile);

    // setup file name buffer and init
    newFilename = strdup(pCurrFile);
    convertCBMString(newFilename, strlen(newFilename));

    pDirEntry = searchDirEntryInDirTrack(pInfo, newFilename, fdLog);
    if (pDirEntry != NULL)
    {
      nextTrack = pDirEntry->firstFileTrack;
      nextSector = pDirEntry->firstFileSector;

      do
      {
        // free next sector
        markSectorAsFree(pInfo, nextTrack, nextSector, fdLog);

        // change to next sector
        rc = GetDiskSectorAddr(pInfo, nextTrack, nextSector, &pAddr);
        if(rc != 0)
        {
          IFDEBUG fprintf(fdLog, "DeleteFiles: bad GetDiskSectorAddr call (%d:%d)\n", nextTrack, nextSector);
          return E_EREAD;
        }
        nextTrack = *pAddr;
        nextSector = *(pAddr+1);

      } while(nextTrack != 0);

      // delete dir entry
      memset(((BYTE *)pDirEntry)+2, 0, sizeof(T_DIR_ENTRY)-2);

      // mark image as changed
      imageChanged = TRUE;
    }
    else
      IFDEBUG fprintf(fdLog, "file <%s> not found in archive\n", newFilename);

    // next file
    free(newFilename);
    pCurrFile += strlen(pCurrFile) + 1;
  }

  IFDEBUG fprintf(fdLog, "deleting done\n");

  if (imageChanged)
  {
    // write image in memory to disk
    fdArchive = fopen(PackedFile, "wb");
    if (fdArchive == NULL)
    {
      IFDEBUG fprintf(fdLog, "error: fopen (write image) failed\n");
      free(pInfo->diskData);
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return E_EWRITE;
    }
    rc = fwrite(pInfo->diskData, 256, 683, fdArchive);  // TODO: !!! 1541 specific
    if (rc != 683)                        // TODO: !!! 1541 specific
    {
      IFDEBUG fprintf(fdLog, "error: fwrite (write image) wrote less (%d != 683)\n", rc);
      free(pInfo->diskData);
      free(pInfo);
      IFDEBUG fclose(fdLog);
      return E_EWRITE;
    }
    fclose(fdArchive);
    IFDEBUG fprintf(fdLog, "image written\n");
  }

  free(pInfo->diskData);
  free(pInfo);

  IFDEBUG fclose(fdLog);

  return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
void __stdcall SetChangeVolProc (HANDLE hArcData, tChangeVolProc pChangeVolProc1)
{
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
void __stdcall SetProcessDataProc (HANDLE hArcData, tProcessDataProc pProcessDataProc)
{
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
void __stdcall PackSetDefaultParams(PackDefaultParamStruct* dps)
{
  FILE *fdLog=NULL;
  char logFile [MAX_PATH];

  determineIniAndLogFile(&fdLog, logFile);
  IFDEBUG fprintf(fdLog, "PackSetDefaultParams\n");
  IFDEBUG fprintf(fdLog, "plugin interface version %d.%d\n", dps->PluginInterfaceVersionHi, dps->PluginInterfaceVersionLow);
  IFDEBUG fprintf(fdLog, "recommended ini file location: %s\n", dps->DefaultIniName);

  IFDEBUG fclose(fdLog);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
#ifdef TESTING
int main (void)
{
  int rc;
  char buffer [1500];
  static char test [] = {
  "15 sec. copy.prg\0"
  "21 sec. copy i.prg\0"
  "9000 monitor.prg\0"
  "austro recomp.prg\0"
  "austrospeed.prg\0"
  "bam corrector.prg\0"
  "bild-irq.prg\0"
  "bordermaker.prg\0"
  "bouncing scroll.prg\0"
  "c000 monitor.prg\0"
  "cfb copy v2.prg\0"
  "cfb copy v3.prg\0"
  "compactor_linker.prg\0"
  "copy +.prg\0"
  "copy 190.prg\0"
  "copy 58.5 (dual).prg\0"
  "copy 58.5.prg\0"
  "cruncher ii.prg\0"
  "disk editor.prg\0"
  "disk marker.prg\0"
  "disk retter.prg\0"
  "disk-wizard 5_86.prg\0"
  "duplicator i.prg\0"
  "duplicator ii.prg\0"
  "e000 copy.prg\0"
  "exbasic level ii.prg\0"
  "fcopy iii 40tr..prg\0"
  "fcopy iii.prg\0"
  "file retter.prg\0"
  "filemaster 40tr..prg\0"
  "files (al).seq\0"
  "files (id).seq\0"
  "fileverfolgung.prg\0"
  "floppy lister.prg\0"
  "game killer.prg\0"
  "game trainer.prg\0"
  "header changer.prg\0"
  "isepic crack.prg\0"
  "j-up speedcopy.prg\0"
  "j64.prg\0"
  "jr intro 1.prg\0"
  "jr intro 2.prg\0"
  "laufschrift i.prg\0"
  "new directory.prg\0"
  "p-basic v.2.prg\0"
  "remkorrector.prg\0"
  "speedcopy i.prg\0"
  "speedcopy ii.prg\0"
  "sprite monitor.prg\0"
  "strich cursor.prg\0"
  "supersorter ii.prg\0"
  "sys -_ run.prg\0"
  "test.prg\0"
  "the viewer.prg\0"
  "toolkit.prg\0"
  "turbo nibbler v4.prg\0"
  "turbo tape iii.prg\0"
};

	memset(buffer, 0, sizeof(buffer));
  //strcpy(buffer, "stub.log");
  //rc = PackFiles("D:\\WORK\\WCX\\dircbm\\stub.d64", NULL, "F:\\", buffer, 0);

  //strcpy(buffer, "spiele_d.csv");
  //rc = PackFiles("D:\\WORK\\WCX\\dircbm\\spiele_c.d64 ", NULL, "D:\\", buffer, 0);

  //strcpy(buffer, "bouncing scroll.prg");
  //rc = PackFiles("D:\\WORK\\WCX\\dircbm\\x.d64 ", NULL, "D:\\WORK\\WCX\\dircbm\\x\\", buffer, 0);

  //rc = PackFiles("D:\\WORK\\WCX\\dircbm\\x.d64 ", NULL, "D:\\WORK\\WCX\\dircbm\\x\\", test, 0);

  //strcpy(buffer, "turbo tape iii.prg");
  //rc = PackFiles("D:\\WORK\\WCX\\dircbm\\x.d64 ", NULL, "D:\\WORK\\WCX\\dircbm\\x\\", buffer, 0);

  //strcpy(buffer, "21 sec. copy i.prg");
  //rc = DeleteFiles("D:\\WORK\\WCX\\dircbm\\x.d64 ", buffer);

  tOpenArchiveData arch;
  tHeaderData header;
  HANDLE hArchive;

//  arch.ArcName = strdup("D:\\WORK\\WCX\\dircbm\\test\\new1\\003_kann d64 nicht ffnen, braucht ziemlich lange\\Electronic Microloops (2004)(CMP).d64");
  arch.ArcName = strdup("D:\\WORK\\WCX\\dircbm\\test\\new1\\000_error-bytes\\Disk310b.d64");
  hArchive = OpenArchive (&arch);
	do
	{
  	rc = ReadHeader (hArchive, &header);
    if (rc != E_END_ARCHIVE)
    {
      printf("found %s\n", header.FileName);
    }
	} while (rc != E_END_ARCHIVE);
}
#endif